/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TableSort.
 *
 * The Initial Developer of the Original Code is
 * Mingyi Liu <mingyiliu@yahoo.com>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 
/* This FF extension derives from my cross-browser html table sorting script
 hosted by on hotscripts.com at http://www.hotscripts.com/Detailed/57817.html
 and on my website (along with demo) at http://www.mingyi.org/other/ts_demo.html.  
 That script can be used in your web pages to display in IE, and it is more 
 powerful */

var ml_tsort = {
  
  init : function()
  {
    if(document.getElementById("contentAreaContextMenu") != null)
    {
      document.getElementById("contentAreaContextMenu").addEventListener("popupshowing", ml_tsort.menuitem_showhide, false);
    }
  },
  
  set_vars : function(event) 
  {
    var ele = event.target;
    ml_tsort.clicked_table = ml_tsort.getParent(ele,'TABLE');
    ml_tsort.clicked_td = ml_tsort.getParent(ele,'TD') || ml_tsort.getParent(ele,'TH');
    ml_tsort.clicked_tr = ml_tsort.getParent(ele,'TR');

    ml_tsort.table_content_might_change = get_pref('tablesort.table_content_might_change', 'bool'); 
    ml_tsort.preserve_style = get_pref('tablesort.preserve_style', 'str'); 
    ml_tsort.tell_me_time_consumption = get_pref('tablesort.tell_me_time_consumption', 'bool'); 
    ml_tsort.default_date_style = get_pref('tablesort.default_date_style', 'str');
    ml_tsort.default_sort_asc = get_pref('tablesort.default_sort_asc', 'bool');
    ml_tsort.alt_click = get_pref('tablesort.alt_click', 'bool');

    ml_tsort.warn_times = parseInt(get_pref('tablesort.warn_times', 'str'));
    if(isNaN(ml_tsort.warn_times) || ml_tsort.warn_times < 0) ml_tsort.warn_times = 0;

    if (event.altKey && event.ctrlKey && ml_tsort.alt_click) ml_tsort.resort_table();
  },
  
  menuitem_showhide : function() 
  {
    if (ml_tsort.clicked_td == null)
    {
      document.getElementById("tsort_menu").hidden = true;
      document.getElementById("tsort_quick_sort").hidden = true;
    }
    else
    {
      document.getElementById("tsort_menu").hidden = false;
      document.getElementById("tsort_quick_sort").hidden = false;
    }
  },
  
  getParent : function(el, pTagName) 
  {
    if (el == null) return null;
    else if (el.nodeType == 1 && el.tagName.toLowerCase() == pTagName.toLowerCase())  // Gecko bug, supposed to be uppercase
      return el;
    else
      return this.getParent(el.parentNode, pTagName);
  },

  getInnerText : function(el) 
  {
    if (typeof el == "string") return el;
    if (typeof el == "undefined") { return el };
    if (el.innerText) return el.innerText;  //Not needed but it is faster
    var str = "";
    
    var cs = el.childNodes;
    var l = cs.length;
    for (var i = 0; i < l; i++) {
      switch (cs[i].nodeType) {
        case 1: //ELEMENT_NODE
          str += this.getInnerText(cs[i]);
          break;
        case 3:  //TEXT_NODE
          str += cs[i].nodeValue;
          break;
      }
    }
    return str;
  },

  match_date_format : function(value, format)
  {
    this.set_date_array(format);
    if(format == 'M/D/Y' && !isNaN(Date.parse(this.getInnerText(value.cells[this.sort_column_index]))))
      return true;
    else if(!isNaN(this.convert_date(value))) return true;
    this.set_date_array(format.replace(/\//g, '-'));
    if(!isNaN(this.convert_date(value))) return true;
    this.set_date_array(format.replace(/\//g, '.'));
    if(!isNaN(this.convert_date(value))) return true;
    this.set_date_array(format.replace(/\//g, ' '));
    if(!isNaN(this.convert_date(value))) return true;
    return false;
  },
    
  resort_table : function(type)
  {
    var td = this.clicked_td;
    if(td == null) return;
    var column = td.cellIndex;
    var table = this.clicked_table;
    this.sort_column_index = column;
    if(table == null || table.rows.length < 1) return;

    if(ml_tsort.warn_times > 0)
    {
      alert("This extension is replaced by the TableTools extension which has more features (it sorts, filters and copies) and bug fixes. Please uninstall this one and install TableTools (https://addons.mozilla.org/firefox/2637/)!\n\nThis message will be displayed " + ml_tsort.warn_times + " more times");
      ml_tsort.warn_times--;
      set_pref('tablesort.warn_times', '' + ml_tsort.warn_times, 'str');
    }

    // now let's do a lot just to save a little time, if possible at all. ;)
    var lastSortCell = table.getAttribute("ts_sortcell") || -1;
    var lastSortType = table.getAttribute("ts_sorttype");
    if(type != null) table.setAttribute("ts_sorttype", type);
    else table.removeAttribute("ts_sorttype");
    var sortingSameColumnAndType = (table == this.last_sorted_table && column == lastSortCell && 
       ((type == null && lastSortType == null) || type == lastSortType));

    var lastSortDir = (sortingSameColumnAndType)? table.getAttribute("ts_sortdir") : ((this.default_sort_asc)? 'desc' : 'asc');
    var newRows = new Array();
    var headcount = 1;
    for (var i=0,j=1;j<table.rows.length;j++)
    {
      var t = table.rows[j].parentNode.tagName.toLowerCase();
      if(t == 'tbody' && table.rows[j].cells.length >= column + 1) newRows[i++] = table.rows[j];
      else if(t == 'thead') headcount++;
    }
    if(newRows.length == 0) return;
    var time2 = new Date();
  
    // check if we really need to sort
    if(!this.table_content_might_change && sortingSameColumnAndType)
      newRows.reverse();
    else
    {
      // Work out a type for the column
      var sortfn;
      this.replace_pattern = '';
      if(type == null)
      {
        var itm;
        for(var i = 0; i < newRows.length; i++)
        {
          itm = this.getInnerText(newRows[i].cells[column]);
          if(itm.match(/\S/)) break;
        }
        if(i == newRows.length) return;
        sortfn = this.sort_caseinsensitive;
        
        // let's try really hard to check dates separated by - or / or .
        if (this.default_date_style == 'euro' && this.match_date_format(newRows[0], 'D/M/Y'))
          sortfn = this.sort_date;
        else if (this.match_date_format(newRows[0], 'M/D/Y')) sortfn = this.sort_date;
        else if (itm.match(/^[$]/)) sortfn = this.sort_currency;
        else if (itm.match(/^\d{1,3}(\.\d{1,3}){3}$/)) sortfn = this.sort_ip;
        else if (itm.match(/^[+-]?\s*[0-9]+(?:\.[0-9]+)?(?:\s*[eE]\s*[+-]?\s*\d+)?$/))
          sortfn = this.sort_numeric;
      }
      else if(type == 'string') sortfn = this.sort_caseinsensitive;
      else if(type == 'date')
      {
        if(this.match_date_format(newRows[0], 'M/D/Y')) sortfn = this.sort_date;
        else alert('The data is not standard time stamp and cannot be parsed! The column will be sorted alphabetically instead.');
      }  
      else if(type == 'euro_date')
      {
        if(this.match_date_format(newRows[0], 'D/M/Y')) sortfn = this.sort_date;
        else alert('The data is not standard time stamp and cannot be parsed! The column will be sorted alphabetically instead.');
      }
      else if(type == 'year_month_date')
      {
        if(this.match_date_format(newRows[0], 'Y/M/D')) sortfn = this.sort_date;
        else alert('The data is not standard time stamp and cannot be parsed! The column will be sorted alphabetically instead.');
      }
      else if(type == 'year_date_month')
      {
        if(this.match_date_format(newRows[0], 'Y/D/M')) sortfn = this.sort_date;
        else alert('The data is not standard time stamp and cannot be parsed! The column will be sorted alphabetically instead.');
      }
      else if(type == 'other_date') 
      { 
        this.set_date_array(td.getAttribute("ts_date_format")); 
        sortfn = this.sort_date; 
      }
      else if(type == 'number') sortfn = this.sort_numeric;
      else if(type == 'ip') sortfn = this.sort_ip;
      else if(type == 'money') sortfn = this.sort_currency;
//       else if(type == 'custom') sortfn = function(aa,bb) { a = this.getInnerText(aa.cells[this.sort_column_index]); b = this.getInnerText(bb.cells[this.sort_column_index]); eval(td.getAttribute("ts_sortfn")) }; // the coding here is shorter but interestingly it's also slower
      else if(type == 'custom') { this.custom_code = td.getAttribute("ts_sortfn"); sortfn = this.custom_sortfn }
      else { alert("unsupported sorting type!"); return; }

      table.setAttribute("ts_sortcell", column);
      newRows.sort(sortfn);
      if (lastSortDir == 'asc') newRows.reverse();
    }
    if (lastSortDir == 'desc') table.setAttribute('ts_sortdir','asc');
    else table.setAttribute('ts_sortdir','desc');
    this.last_sorted_table = table;
       
    var time3 = new Date();
    
    if(this.preserve_style == 'row')
    {
      var tmp = new Array(newRows.length);
      for (var i = 0; i < newRows.length; i++) tmp[i] = newRows[i].innerHTML;
      for (var i = 0; i < newRows.length; i++) table.rows[i+headcount].innerHTML = tmp[i];
    }
    else if(this.preserve_style == 'cell')
    {
      var tmp = new Array(newRows.length);
      for (var i = 0; i < newRows.length; i++)
        for (var j = 0; j < newRows[i].cells.length; j++)
        {
          if(!tmp[i]) tmp[i] = new Array(newRows[i].cells.length);
          tmp[i][j] = newRows[i].cells[j].innerHTML;
        }
      for (var i = 0; i < newRows.length; i++)
        for (var j = 0; j < newRows[i].cells.length; j++)
          table.rows[i+headcount].cells[j].innerHTML = tmp[i][j];
    }
    else
    {
      for (var i=0;i<newRows.length;i++) // We appendChild rows that already exist to the tbody, so it moves them rather than creating new ones
        table.tBodies[0].appendChild(newRows[i]);
    } 
    var time4 = new Date();
    if(this.tell_me_time_consumption)
    {
      alert('it took ' + this.diff_time(time3, time2) + ' seconds to do sorting!');
      alert('it took ' + this.diff_time(time4, time3) + ' seconds to do redrawing!');
    }
  },
  
  diff_time : function(time2, time1) 
  {
    return (time2.getTime() - time1.getTime())/1000;
  },
  
  // Mingyi Note: it seems ridiculous to do so much processing for
  // customizable date conversion, should try to find a zbetter way
  // of doing it.
  set_date_array : function(f) 
  {
    var tmp = [['D', f.indexOf('D')], ['M', f.indexOf('M')], ['Y', f.indexOf('Y')]];
    tmp.sort(function(a,b){ return a[1] - b[1]});
    this.date_order_array = new Array(3);
    for(var i = 0; i < 3; i++) this.date_order_array[tmp[i][0]] = '$' + (i + 2);
    this.replace_pattern = f.replace(/[DMY]([^DMY]+)[DMY]([^DMY]+)[DMY]/, '^(.*?)(\\d+)\\$1(\\d+)\\$2(\\d+)(.*)$');
  },
  
  process_year : function(y) 
  {
    var tmp = parseInt(y);
    if(tmp < 32) return '20' + y; 
    else if(tmp < 100) return '19' + y;
    else return y;
  },

  // convert to MM/DD/YYYY (or M/D/YYYY) format
  convert_date : function(a) 
  {
    var aa = this.getInnerText(a.cells[this.sort_column_index]);
    var re = 'RegExp.$1+RegExp.'+this.date_order_array['M']+'+\'/\'+RegExp.'+this.date_order_array['D']+'+\'/\'+this.process_year(RegExp.'+this.date_order_array['Y']+')+RegExp.$5';
    var code = 'if(aa.match(/'+this.replace_pattern+'/)) (' + re + ')';
    return Date.parse(eval(code));
  },
  
  sort_date : function(a,b) 
  {
    var aa, bb;
    // basically I have to do the conversion due to the potential usage of double digit years
    aa = ml_tsort.convert_date(a);
    if(isNaN(aa)) aa = Date.parse(ml_tsort.getInnerText(a.cells[ml_tsort.sort_column_index]));
    bb = ml_tsort.convert_date(b);
    if(isNaN(bb)) bb = Date.parse(ml_tsort.getInnerText(b.cells[ml_tsort.sort_column_index]));
    if(isNaN(aa)) aa = 0;
    if(isNaN(bb)) bb = 0;
    return aa - bb;
  },
  
  // assume no scientific number in currency (if assumption incorrect, just use
  // same code for this.sort_numeric will do)
  sort_currency : function(a,b) 
  { 
    return ml_tsort.sort_num(ml_tsort.getInnerText(a.cells[ml_tsort.sort_column_index]).replace(/[^-0-9.+]/g,''),
                       ml_tsort.getInnerText(b.cells[ml_tsort.sort_column_index]).replace(/[^-0-9.+]/g,''));
  },
  
  // let's allow scientific notation but also be strict on number format
  sort_num : function(a, b) 
  {
    var aa, bb;
    if(!isNaN(a)) aa = a;
    else if(a && a.match(/^[^0-9.+-]*([+-]?\s*[0-9]+(?:\.[0-9]+)?(?:\s*[eE]\s*[+-]?\s*\d+)?)/))
      aa = parseFloat(RegExp.$1.replace(/\s+/g, ''));
    else aa = 0;
    if(!isNaN(b)) bb = b;
    else if(b && b.match(/^[^0-9.+-]*([+-]?\s*[0-9]+(?:\.[0-9]+)?(?:\s*[eE]\s*[+-]?\s*\d+)?)/))
      bb = parseFloat(RegExp.$1.replace(/\s+/g, ''));
    else bb = 0;
    return aa - bb;
  },
  
  sort_numeric : function(a,b) 
  {
    return ml_tsort.sort_num(ml_tsort.getInnerText(a.cells[ml_tsort.sort_column_index]),
                       ml_tsort.getInnerText(b.cells[ml_tsort.sort_column_index])); 
  },
  
  sort_ip : function(a,b) 
  {
    var aa = ml_tsort.getInnerText(a.cells[ml_tsort.sort_column_index]).split('.');
    var bb = ml_tsort.getInnerText(b.cells[ml_tsort.sort_column_index]).split('.');
    return ml_tsort.sort_num(aa[0], bb[0]) || ml_tsort.sort_num(aa[1], bb[1]) || 
           ml_tsort.sort_num(aa[2], bb[2]) || ml_tsort.sort_num(aa[3], bb[3]);
  },
   
  sort_caseinsensitive : function(a,b) 
  {
    var aa = ml_tsort.getInnerText(a.cells[ml_tsort.sort_column_index]).toLowerCase();
    var bb = ml_tsort.getInnerText(b.cells[ml_tsort.sort_column_index]).toLowerCase();
    if (aa==bb) return 0;
    if (aa<bb) return -1;
    return 1;
  },
  
  custom_sortfn : function(aa,bb) 
  {
    var a = ml_tsort.getInnerText(aa.cells[ml_tsort.sort_column_index]);
    var b = ml_tsort.getInnerText(bb.cells[ml_tsort.sort_column_index]);
    return eval(ml_tsort.custom_code);
  }
};

function get_pref(key, type)
{
  var p = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
  if(!type || type == 'string' || type == 'str') return p.getCharPref(key);
  if(type == 'boolean' || type == 'bool') return p.getBoolPref(key);  
}

function set_pref(key, value, type)
{
  var p = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
  if(!type || type == 'string' || type == 'str') p.setCharPref(key, value);
  if(type == 'boolean' || type == 'bool') p.setBoolPref(key, value);
}

window.addEventListener("click", ml_tsort.set_vars, false);
window.addEventListener("load", ml_tsort.init, false);
